searchData={"items":[{"type":"module","title":"MuonTrap","doc":"MuonTrap protects you from lost and out of control OS processes.\n\nYou can use it as a `System.cmd/3` replacement or to pull OS processes into\nan Erlang supervision tree via `MuonTrap.Daemon`. Either way, if the Erlang\nprocess that runs the command dies, then the OS processes will die as well.\n\nMuonTrap tries very hard to kill OS processes so that remnants don't hang\naround the system when your Erlang code thinks they should be gone. MuonTrap\ncan use the Linux kernel's `cgroup` feature to contain the child process and\nall of its children. From there, you can limit CPU and memory and other\nresources to the process group.\n\nMuonTrap does not require `cgroups` but keep in mind that OS processes can\nescape. It is, however, still an improvement over `System.cmd/3` which does\nnot have a mechanism for dealing it OS processes that do not monitor their\nstdin for when to close.\n\nFor more information, see the documentation for `MuonTrap.cmd/3` and\n`MuonTrap.Daemon`","ref":"MuonTrap.html"},{"type":"module","title":"Configuring cgroups - MuonTrap","doc":"On most Linux distributions, use `cgcreate` to create a new cgroup.  You can\nname them almost anything. The command below creates one named `muontrap` for\nthe current user. It supports memory and CPU controls.\n\n```sh\nsudo cgcreate -a $(whoami) -g memory,cpu:muontrap\n```\n\nNerves systems do not contain `cgcreate` by default. Due to the simpler Linux\nsetup, it may be sufficient to run `File.mkdir_p(cgroup_path)` to create a\ncgroup. For example:\n\n```elixir\nFile.mkdir_p(\"/sys/fs/cgroup/memory/muontrap\")\n```\n\nThis creates the cgroup path, `muontrap` under the `memory` controller.  If\nyou do not have the `\"/sys/fs/cgroup\"` directory, you will need to mount it\nor update your `erlinit.config` to mount it for you. See a newer official\nsystem for an example.","ref":"MuonTrap.html#module-configuring-cgroups"},{"type":"function","title":"MuonTrap.cmd/3","doc":"Executes a command like `System.cmd/3` via the `muontrap` wrapper.","ref":"MuonTrap.html#cmd/3"},{"type":"function","title":"Options - MuonTrap.cmd/3","doc":"* `:cgroup_controllers` - run the command under the specified cgroup controllers. Defaults to `[]`.\n  * `:cgroup_base` - create a temporary path under the specified cgroup path\n  * `:cgroup_path` - explicitly specify a path to use. Use `:cgroup_base`, unless you must control the path.\n  * `:cgroup_sets` - set a cgroup controller parameter before running the command\n  * `:delay_to_sigkill` - milliseconds before sending a SIGKILL to a child process if it doesn't exit with a SIGTERM (default 500 ms)\n  * `:uid` - run the command using the specified uid or username\n  * `:gid` - run the command using the specified gid or group\n  * `:timeout` - milliseconds to wait for the command to complete. If the\n    command does not exit before the timeout, the return value will contain\n    the output up to that point and `:timeout` as the exit status. The child\n    process will be sent SIGTERM\n\nThe following `System.cmd/3` options are also available:\n\n  * `:into` - injects the result into the given collectable, defaults to `\"\"`\n  * `:cd` - the directory to run the command in\n  * `:env` - an enumerable of tuples containing environment key-value as binary\n  * `:arg0` - sets the command arg0\n  * `:stderr_to_stdout` - redirects stderr to stdout when `true`\n  * `:parallelism` - when `true`, the VM will schedule port tasks to improve\n    parallelism in the system. If set to `false`, the VM will try to perform\n    commands immediately, improving latency at the expense of parallelism.\n    The default can be set on system startup by passing the \"+spp\" argument\n    to `--erl`.","ref":"MuonTrap.html#cmd/3-options"},{"type":"function","title":"Examples - MuonTrap.cmd/3","doc":"Run a command:\n\n```elixir\niex> MuonTrap.cmd(\"echo\", [\"hello\"])\n{\"hello\\n\", 0}\n```\n\nThe next examples only run on Linux. To try this out, create new cgroups:\n\n```sh\nsudo cgcreate -a $(whoami) -g memory,cpu:muontrap\n```\n\nRun a command, but limit memory so severely that it doesn't work (for demo\npurposes, obviously):\n\n```elixir\niex-donttest> MuonTrap.cmd(\"echo\", [\"hello\"], cgroup_controllers: [\"memory\"], cgroup_path: \"muontrap/test\", cgroup_sets: [{\"memory\", \"memory.limit_in_bytes\", \"8192\"}])\n{\"\", 1}\n```\n\nRun a command with a timeout:\n\niex> MuonTrap.cmd(\"/bin/sh\", [\"-c\", \"echo start && sleep 10 && echo end\"], timeout: 100)\n{\"start\\n\", :timeout}","ref":"MuonTrap.html#cmd/3-examples"},{"type":"function","title":"MuonTrap.muontrap_path/0","doc":"Return the absolute path to the muontrap executable.\n\nCall this if you want to invoke the `muontrap` port binary manually.","ref":"MuonTrap.html#muontrap_path/0"},{"type":"module","title":"MuonTrap.Daemon","doc":"Wrap an OS process in a GenServer so that it can be supervised.\n\nFor example, in your children list add MuonTrap.Daemon like this:\n\n```elixir\nchildren = [\n  {MuonTrap.Daemon, [\"my_server\", [\"--options\", \"foo\"]], [cd: \"/some_directory\"]]}\n]\n\nopts = [strategy: :one_for_one, name: MyApplication.Supervisor]\nSupervisor.start_link(children, opts)\n```\n\nIn the `child_spec` tuple, the second element is a list that corresponds to\nthe `MuonTrap.cmd/3` parameters. I.e., The first item in the list is the\nprogram to run, the second is a list of commandline arguments, and the third\nis a list of options. The same options as `MuonTrap.cmd/3` are available with\nthe following additions:\n\n* `:name` - Name the Daemon GenServer\n* `:log_output` - When set, send output from the command to the Logger.\n  Specify the log level (e.g., `:debug`)\n* `:log_prefix` - Prefix each log message with this string (defaults to the\n  program's path)\n* `:log_transform` - Pass a function that takes a string and returns a string\n  to format output from the command. Defaults to `String.replace_invalid/1`\n  on Elixir 1.16+ to avoid crashing the logger on non-UTF8 output.\n* `:logger_metadata` - A keyword list to merge into the process's logger metadata.\n  The `:muontrap_cmd` and `:muontrap_args` keys are automatically added and\n  cannot be overridden.\n* `:stderr_to_stdout` - When set to `true`, redirect stderr to stdout.\n  Defaults to `false`.\n* `:exit_status_to_reason` - Optional function to convert the exit status (a\n  number) to stop reason for the Daemon GenServer. Use if error exit codes\n  carry information or aren't errors.\n\nIf you want to run multiple `MuonTrap.Daemon`s under one supervisor, they'll\nall need unique IDs. Use `Supervisor.child_spec/2` like this:\n\n```elixir\nSupervisor.child_spec({MuonTrap.Daemon, [\"my_server\", []]}, id: :server1)\n```","ref":"MuonTrap.Daemon.html"},{"type":"function","title":"MuonTrap.Daemon.cgget/3","doc":"Get the value of the specified cgroup variable.","ref":"MuonTrap.Daemon.html#cgget/3"},{"type":"function","title":"MuonTrap.Daemon.cgset/4","doc":"Modify a cgroup variable.","ref":"MuonTrap.Daemon.html#cgset/4"},{"type":"function","title":"MuonTrap.Daemon.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"MuonTrap.Daemon.html#child_spec/1"},{"type":"function","title":"MuonTrap.Daemon.os_pid/1","doc":"Return the OS pid to the muontrap executable.","ref":"MuonTrap.Daemon.html#os_pid/1"},{"type":"function","title":"MuonTrap.Daemon.start_link/3","doc":"Start/link a deamon GenServer for the specified command.","ref":"MuonTrap.Daemon.html#start_link/3"},{"type":"function","title":"MuonTrap.Daemon.statistics/1","doc":"Return statistics about the daemon\n\nStatistics:\n\n* `:output_byte_count` - bytes output by the process being run","ref":"MuonTrap.Daemon.html#statistics/1"},{"type":"module","title":"MuonTrap.Options","doc":"Validate and normalize the options passed to MuonTrap.cmd/3 and MuonTrap.Daemon.start_link/3\n\nThis module is generally not called directly, but it's likely\nthe source of exceptions if any options aren't quite right. Call `validate/4` directly to\ndebug or check options without invoking a command.","ref":"MuonTrap.Options.html"},{"type":"function","title":"MuonTrap.Options.validate/4","doc":"Validate options and normalize them for invoking commands\n\nPass in `:cmd` or `:daemon` for the first parameter to allow function-specific\noptions.","ref":"MuonTrap.Options.html#validate/4"},{"type":"type","title":"MuonTrap.Options.t/0","doc":"The following fields are always present:\n\n* `:cmd` - the command to run\n* `:args` - a list of arguments to the command\n\nThe next fields are optional:\n\n* `:into` - `MuonTrap.cmd/3` only\n* `:cd`\n* `:arg0`\n* `:stderr_to_stdout`\n* `:parallelism`\n* `:env`\n* `:name` - `MuonTrap.Daemon`-only\n* `:log_output` - `MuonTrap.Daemon`-only\n* `:log_prefix` - `MuonTrap.Daemon`-only\n* `:log_transform` - `MuonTrap.Daemon`-only\n* `:logger_metadata` - `MuonTrap.Daemon`-only\n* `:stdio_window`\n* `:exit_status_to_reason` - `MuonTrap.Daemon`-only\n* `:cgroup_controllers`\n* `:cgroup_path`\n* `:cgroup_base`\n* `:delay_to_sigkill`\n* `:cgroup_sets`\n* `:uid`\n* `:gid`\n* `:timeout` - `MuonTrap.cmd/3` only","ref":"MuonTrap.Options.html#t:t/0"},{"type":"extras","title":"MuonTrap","doc":"# MuonTrap\n\n[![CircleCI](https://dl.circleci.com/status-badge/img/gh/fhunleth/muontrap/tree/main.svg?style=svg)](https://dl.circleci.com/status-badge/redirect/gh/fhunleth/muontrap/tree/main)\n[![Hex version](https://img.shields.io/hexpm/v/muontrap.svg \"Hex version\")](https://hex.pm/packages/muontrap)\n[![REUSE status](https://api.reuse.software/badge/github.com/fhunleth/muontrap)](https://api.reuse.software/info/github.com/fhunleth/muontrap)\n\nKeep programs, deamons, and applications launched from Erlang and Elixir\ncontained and well-behaved. This lightweight library kills OS processes if the\nElixir process running them crashes and if you're running on Linux, it can use\ncgroups to prevent many other shenanigans.\n\nSome other features:\n\n* Attach your OS process to a supervision tree via a convenient `child_spec`\n* Set `cgroup` controls like thresholds on memory and CPU utilization\n* Start OS processes as a different user or group\n* Send SIGKILL to processes that aren't responsive to SIGTERM\n* With `cgroups`, ensure that all children of launched processes have been killed too","ref":"readme.html"},{"type":"extras","title":"TL;DR - MuonTrap","doc":"Add `muontrap` to your project's `mix.exs` dependency list:\n\n```elixir\ndef deps do\n  [\n    {:muontrap, \"~> 1.0\"}\n  ]\nend\n```\n\nRun a command similar to\n[`System.cmd/3`](https://hexdocs.pm/elixir/System.html#cmd/3):\n\n```elixir\niex>  MuonTrap.cmd(\"echo\", [\"hello\"])\n{\"hello\\n\", 0}\n```\n\nAttach a long running process to a supervision tree using a\n[child_spec](https://hexdocs.pm/elixir/Supervisor.html#module-child-specification)\nlike the following:\n\n```elixir\n{MuonTrap.Daemon, [\"long_running_command\", [\"arg1\", \"arg2\"], options]}\n```\n\nRunning on Linux and can use cgroups? Then create a new cgroup:\n\n```bash\nsudo cgcreate -a $(whoami) -g memory:mycgroup\n```\n\n```elixir\n{MuonTrap.Daemon,\n [\n   \"long_running_command\",\n   [\"arg1\", \"arg2\"],\n   [cgroup_controllers: [\"memory\"], cgroup_base: \"mycgroup\"]\n ]}\n```\n\n`MuonTrap` will create a cgroup under \"mycgroup\" to run the\n`\"long_running_command\"`. If the command fails, it will be restarted. If it\nshould no longer be running (like if something else crashed in Elixir and\nsupervision needs to clean up) then MuonTrap will kill `\"long_running_command\"`\nand all of its children.\n\nWant to know more? Read on...","ref":"readme.html#tl-dr"},{"type":"extras","title":"The problem - MuonTrap","doc":"The Erlang VM's port interface lets Elixir applications run external programs.\nThis is important since it's not practical to rewrite everything in Elixir.\nPlus, if the program is long running like a daemon or a server, you use Elixir\nto supervise it and restart it on crashes. The catch is that the Erlang VM\nexpects port processes to be well-behaved. As you'd expect, many useful programs\ndon't quite meet the Erlang VM's expectations.\n\nFor example, let's say that you want to monitor a network connection and decide\nthat `ping` is the right tool. Here's how you could start `ping` in a process.\n\n```elixir\niex> pid = spawn(fn -> System.cmd(\"ping\", [\"-i\", \"5\", \"localhost\"], into: IO.stream(:stdio, :line)) end)\n#PID<0.6116.0>\nPING localhost (127.0.0.1): 56 data bytes\n64 bytes from 127.0.0.1: icmp_seq=0 ttl=64 time=0.032 ms\n64 bytes from 127.0.0.1: icmp_seq=1 ttl=64 time=0.077 ms\n```\n\nTo see that `ping` is running, call `ps` to look for it. You can also do this\nfrom a separate terminal window outside of IEx:\n\n```elixir\niex> :os.cmd('ps -ef | grep ping') |> IO.puts\n  501 38820 38587   0  9:26PM ??         0:00.01 /sbin/ping -i 5 localhost\n  501 38824 38822   0  9:27PM ??         0:00.00 grep ping\n:ok\n```\n\nNow exit the Elixir process. Imagine here that in the real program that\nsomething happened in Elixir and the process needs to exit and be restarted by a\nsupervisor.\n\n```elixir\niex> Process.exit(pid, :oops)\ntrue\niex> :os.cmd('ps -ef | grep ping') |> IO.puts\n  501 38820 38587   0  9:26PM ??         0:00.02 /sbin/ping -i 5 localhost\n  501 38833 38831   0  9:34PM ??         0:00.00 grep ping\n```\n\nAs you can tell, `ping` is still running after the exit. If you run `:observer`\nyou'll see that Elixir did indeed terminate both the process and the port, but\nthat didn't stop `ping`. The reason for this is that `ping` doesn't pay\nattention to `stdin` and doesn't notice the Erlang VM closing it to signal that\nit should exit.\n\nImagine now that the process was supervised and it restarts. If this happens a\nregularly, you could be running dozens of `ping` commands.\n\nThis is just one of the problems that `muontrap` fixes.","ref":"readme.html#the-problem"},{"type":"extras","title":"Applicability - MuonTrap","doc":"This is intended for long running processes. It's not great for interactive\nprograms that communicate via the port or send signals. That feature is possible\nto add, but you'll probably be happier with other solutions like\n[erlexec](https://github.com/saleyn/erlexec/).","ref":"readme.html#applicability"},{"type":"extras","title":"Running commands - MuonTrap","doc":"The simplest way to use `muontrap` is as a replacement to `System.cmd/3`. Here's\nan example using `ping`:\n\n```elixir\niex> pid = spawn(fn -> MuonTrap.cmd(\"ping\", [\"-i\", \"5\", \"localhost\"], into: IO.stream(:stdio, :line)) end)\n#PID<0.30860.0>\nPING localhost (127.0.0.1): 56 data bytes\n64 bytes from 127.0.0.1: icmp_seq=0 ttl=64 time=0.027 ms\n64 bytes from 127.0.0.1: icmp_seq=1 ttl=64 time=0.081 ms\n```\n\nNow if you exit that process, `ping` gets killed as well:\n\n```elixir\niex> Process.exit(pid, :oops)\ntrue\niex> :os.cmd('ps -ef | grep ping') |> IO.puts\n  501 38898 38896   0  9:58PM ??         0:00.00 grep ping\n\n:ok\n```","ref":"readme.html#running-commands"},{"type":"extras","title":"Containment with cgroups - MuonTrap","doc":"Even if you don't make use of any cgroup controller features, having your port\nprocess contained can be useful just to make sure that everything is cleaned\nup on exit including any subprocesses.\n\nTo set this up, first create a cgroup with appropriate permissions. Any path\nwill do; `muontrap` just needs to be able to create a subdirectory underneath it\nfor its use. For example:\n\n```bash\nsudo cgcreate -a $(whoami) -g memory,cpu:mycgroup\n```\n\nBe sure to create the group for all of the cgroup controllers that you wish to\nuse with `muontrap`. The above example creates it for the `memory` and `cpu`\ncontrollers.\n\nIn Elixir, call `MuonTrap.cmd/3` with the\ncgroup options now. In this case, we'll use the `cpu` controller, but this\nexample would work fine with any of the controllers.\n\n```elixir\niex>  MuonTrap.cmd(\"spawning_program\", [], cgroup_controllers: [\"cpu\"], cgroup_base: \"mycgroup\")\n{\"hello\\n\", 0}\n```\n\nIn this example, `muontrap` runs `spawning_program` in a sub-cgroup under the\n`cpu/mycgroup` group. The cgroup parameters may be modified outside of\n`muontrap` using `cgset` or my accessing the cgroup mountpoint manually.\n\nOn any error or if the Erlang VM closes the port or if `spawning_program` exits,\n`muontrap` will kill all OS processes in cgroup. No need to worry about\nrandom processes accumulating on your system.\n\nNote that if you use `cgroup_base`, a temporary cgroup is created for running\nthe command. If you want `muontrap` to use a particular cgroup and not create a\nsubgroup for the command, use the `:cgroup_path` option. Note that if you\nexplicitly specify a cgroup, be careful not to use it for anything else.\n`MuonTrap` assumes that it owns the cgroup and when it needs to kill processes,\nit kills all of them in the cgroup.\n\n#","ref":"readme.html#containment-with-cgroups"},{"type":"extras","title":"Limit the memory used by a process - MuonTrap","doc":"Linux's cgroups are very powerful and the examples here only scratch the\nsurface. If you'd like to limit an OS process and all of its child processes to\na maximum amount of memory, you can do that with the `memory` controller:\n\n```elixir\niex>  MuonTrap.cmd(\"memory_hog\", [], cgroup_controllers: [\"memory\"], cgroup_base: \"mycgroup\", cgroup_sets: [{\"memory\", \"memory.limit_in_bytes\", \"268435456\"}])\n```\n\nThat line restricts the total memory used by `memory_hog` to 256 MB.\n\n#","ref":"readme.html#limit-the-memory-used-by-a-process"},{"type":"extras","title":"Limit CPU usage in a port - MuonTrap","doc":"Limiting the maximum CPU usage is also possible. Two parameters control that\nwith the `cpu` controller: `cpu.cfs_period_us` specifies the number of\nmicroseconds in the scheduling period and `cpu.cfs_quota_us` specifies how many\nof those microseconds can be used. Here's an example call that prevents a\nprogram from using more than 50% of the CPU:\n\n```elixir\niex>  MuonTrap.cmd(\"cpu_hog\", [], cgroup_controllers: [\"cpu\"], cgroup_base: \"mycgroup\", cgroup_sets: [{\"cpu\", \"cpu.cfs_period_us\", \"100000\"}, {\"cpu\", \"cpu.cfs_quota_us\", 50000}])\n```","ref":"readme.html#limit-cpu-usage-in-a-port"},{"type":"extras","title":"Supervision - MuonTrap","doc":"For many long running programs, you may want to restart them if they crash.\nLuckily Erlang already has mechanisms to do this. `MuonTrap` provides a\n`GenServer` called `MuonTrap.Daemon` that you can hook into one of your\nsupervision trees.  For example, you could specify it like this in your\napplication's supervisor:\n\n```elixir\n  def start(_type, _args) do\n    children = [\n      {MuonTrap.Daemon, [\"command\", [\"arg1\", \"arg2\"], options]}\n    ]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n```\n\nSupervisors provide three restart strategies, `:permanent`, `:temporary`, and\n`:transient`. They work as follows:\n\n* `:permanent` - Always restart the command if it exits or crashes. Restarts are\n  limited to the Supervisor's restart intensity settings as they would be with\n  normal `GenServer`s. This is the default.\n* `:transient` - If the exit status of the command is 0 (i.e., success), then\n  don't restart. Any other exit status is considered an error and the command is\n  restarted.\n* `:temporary` - Don't restart\n\nIf you're running more than one `MuonTrap.Daemon` under the same `Supervisor`,\nthen you'll need to give each one a unique `:id`. Here's an example `child_spec`\nfor setting the `:id` and the `:restart` parameters:\n\n```elixir\n    Supervisor.child_spec(\n        {MuonTrap.Daemon, [\"command\", [\"arg1\"], options]},\n         id: :my_daemon,\n         restart: :transient\n      )\n```","ref":"readme.html#supervision"},{"type":"extras","title":"stdio flow control - MuonTrap","doc":"The Erlang port feature does not implement flow control from messages coming\nfrom the port process. Since `MuonTrap` captures stdio from the program being\nrun, it's possible that the program sends output so fast that it grows the\nElixir process's mailbox big enough to cause an out-of-memory error.\n\n`MuonTrap` protects against this by implementing a flow control mechanism. When\ntriggered, the running program's stdout and stderr file handles won't be read\nand hence it will eventually be blocked from writing to those handles.\n\nThe `:stdio_window` option specifies the maximum number of unacknowledged bytes\nallowed. The default is 10 KB.","ref":"readme.html#stdio-flow-control"},{"type":"extras","title":"muontrap development - MuonTrap","doc":"In order to run the tests, some additional tools need to be installed.\nSpecifically the `cgcreate` and `cgget` binaries need to be installed (and\navailable on `$PATH`). Typically the package may be called `cgroup-tools` (on\narch linux you need to install the `libcgroup` aur package).\n\nThen run:\n\n```sh\nsudo cgcreate -a $(whoami) -g memory,cpu:muontrap_test\n```","ref":"readme.html#muontrap-development"},{"type":"extras","title":"License - MuonTrap","doc":"All original source code in this project is licensed under Apache-2.0.\n\nAdditionally, this project follows the [REUSE recommendations](https://reuse.software)\nand labels so that licensing and copyright are clear at the file level.\n\nExceptions to Apache-2.0 licensing are:\n\n* Configuration and data files are licensed under CC0-1.0\n* Documentation is CC-BY-4.0","ref":"readme.html#license"}],"content_type":"text/markdown"}